@@ -1,7 +1,7 @@
 #ifndef CLIENT_COMMON_H
 #define CLIENT_COMMON_H
 
-#include "network.h"
+#include "net.h"
 #include "login-proxy.h"
 #include "sasl-server.h"
 #include "master-login.h" /* for LOGIN_MAX_SESSION_ID_LEN */
@@ -33,29 +33,37 @@
 #define AUTH_MASTER_WAITING_MSG \
 	"Waiting for authentication master process to respond.."
 
-enum client_cmd_reply {
-	CLIENT_CMD_REPLY_OK,
-	CLIENT_CMD_REPLY_AUTH_FAILED,
-	CLIENT_CMD_REPLY_AUTHZ_FAILED,
-	CLIENT_CMD_REPLY_AUTH_FAIL_TEMP,
-	CLIENT_CMD_REPLY_AUTH_FAIL_REASON,
-	CLIENT_CMD_REPLY_AUTH_FAIL_NOSSL,
-	CLIENT_CMD_REPLY_BAD,
-	CLIENT_CMD_REPLY_BYE,
-	CLIENT_CMD_REPLY_STATUS,
-	CLIENT_CMD_REPLY_STATUS_BAD
+enum client_disconnect_reason {
+	CLIENT_DISCONNECT_TIMEOUT,
+	CLIENT_DISCONNECT_SYSTEM_SHUTDOWN,
+	CLIENT_DISCONNECT_RESOURCE_CONSTRAINT,
+	CLIENT_DISCONNECT_INTERNAL_ERROR
+};
+
+enum client_auth_result {
+	CLIENT_AUTH_RESULT_SUCCESS,
+	CLIENT_AUTH_RESULT_REFERRAL_SUCCESS,
+	CLIENT_AUTH_RESULT_REFERRAL_NOLOGIN,
+	CLIENT_AUTH_RESULT_ABORTED,
+	CLIENT_AUTH_RESULT_AUTHFAILED,
+	CLIENT_AUTH_RESULT_AUTHFAILED_REASON,
+	CLIENT_AUTH_RESULT_AUTHZFAILED,
+	CLIENT_AUTH_RESULT_TEMPFAIL,
+	CLIENT_AUTH_RESULT_SSL_REQUIRED
 };
 
 struct client_auth_reply {
 	const char *master_user, *reason;
 	/* for proxying */
-	const char *host, *hostip, *destuser, *password;
+	const char *host, *hostip, *source_ip;
+	const char *destuser, *password, *proxy_mech;
 	unsigned int port;
 	unsigned int proxy_timeout_msecs;
 	unsigned int proxy_refresh_secs;
 	enum login_proxy_ssl_flags ssl_flags;
 
 	unsigned int proxy:1;
+	unsigned int proxy_nopipelining:1;
 	unsigned int temp:1;
 	unsigned int nologin:1;
 	unsigned int authz_failure:1;
@@ -65,17 +73,25 @@ struct client_vfuncs {
 	struct client *(*alloc)(pool_t pool);
 	void (*create)(struct client *client, void **other_sets);
 	void (*destroy)(struct client *client);
-	void (*send_greeting)(struct client *client);
+	void (*notify_auth_ready)(struct client *client);
+	void (*notify_disconnect)(struct client *client,
+				  enum client_disconnect_reason reason,
+				  const char *text);
+	void (*notify_status)(struct client *client,
+			      bool bad, const char *text);
+	void (*notify_starttls)(struct client *client,
+				bool success, const char *text);
 	void (*starttls)(struct client *client);
 	void (*input)(struct client *client);
-	void (*send_line)(struct client *client, enum client_cmd_reply reply,
-			  const char *text);
-	bool (*auth_handle_reply)(struct client *client,
-				  const struct client_auth_reply *reply);
 	void (*auth_send_challenge)(struct client *client, const char *data);
-	int (*auth_parse_response)(struct client *client);
+	void (*auth_parse_response)(struct client *client);
+	void (*auth_result)(struct client *client,
+			    enum client_auth_result result,
+			    const struct client_auth_reply *reply,
+			    const char *text);
 	void (*proxy_reset)(struct client *client);
 	int (*proxy_parse_line)(struct client *client, const char *line);
+	void (*proxy_error)(struct client *client, const char *text);
 };
 
 struct client {
@@ -88,10 +104,12 @@ struct client {
 
 	struct ip_addr local_ip;
 	struct ip_addr ip;
-	struct ip_addr real_ip;
+	struct ip_addr real_remote_ip, real_local_ip;
 	unsigned int local_port, remote_port;
+	unsigned int real_local_port, real_remote_port;
 	struct ssl_proxy *ssl_proxy;
 	const struct login_settings *set;
+	const struct master_service_ssl_settings *ssl_set;
 	const char *session_id;
 
 	int fd;
@@ -106,7 +124,10 @@ struct client {
 
 	struct login_proxy *login_proxy;
 	char *proxy_user, *proxy_master_user, *proxy_password;
+	const struct dsasl_client_mech *proxy_mech;
+	struct dsasl_client *proxy_sasl_client;
 	unsigned int proxy_state;
+	unsigned int proxy_ttl;
 
 	char *auth_mech_name;
 	struct auth_client_request *auth_request;
@@ -122,11 +143,10 @@ struct client {
 	unsigned int auth_attempts, auth_successes;
 	pid_t mail_pid;
 
-	char *virtual_user;
+	char *virtual_user, *virtual_user_orig, *virtual_auth_user;
 	unsigned int destroyed:1;
 	unsigned int input_blocked:1;
 	unsigned int login_success:1;
-	unsigned int greeting_sent:1;
 	unsigned int starttls:1;
 	unsigned int tls:1;
 	unsigned int secured:1;
@@ -139,9 +159,12 @@ struct client {
 	unsigned int auth_initializing:1;
 	unsigned int auth_process_comm_fail:1;
 	unsigned int proxy_auth_failed:1;
+	unsigned int proxy_nopipelining:1;
 	unsigned int auth_waiting:1;
 	unsigned int auth_user_disabled:1;
 	unsigned int auth_pass_expired:1;
+	unsigned int notified_auth_ready:1;
+	unsigned int notified_disconnect:1;
 	/* ... */
 };
 
@@ -149,14 +172,16 @@ extern struct client *clients;
 
 struct client *
 client_create(int fd, bool ssl, pool_t pool,
-	      const struct login_settings *set, void **other_sets,
+	      const struct login_settings *set,
+	      const struct master_service_ssl_settings *ssl_set,
+	      void **other_sets,
 	      const struct ip_addr *local_ip, const struct ip_addr *remote_ip);
 void client_destroy(struct client *client, const char *reason);
 void client_destroy_success(struct client *client, const char *reason);
 void client_destroy_internal_failure(struct client *client);
 
 void client_ref(struct client *client);
-bool client_unref(struct client **client);
+bool client_unref(struct client **client) ATTR_NOWARN_UNUSED_RESULT;
 
 void client_cmd_starttls(struct client *client);
 
@@ -167,22 +192,28 @@ void client_log(struct client *client, const char *msg);
 void client_log_err(struct client *client, const char *msg);
 void client_log_warn(struct client *client, const char *msg);
 const char *client_get_extra_disconnect_reason(struct client *client);
-bool client_is_trusted(struct client *client);
-void client_auth_failed(struct client *client);
+
+void client_auth_respond(struct client *client, const char *response);
+void client_auth_abort(struct client *client);
 bool client_is_tls_enabled(struct client *client);
+void client_auth_fail(struct client *client, const char *text);
 const char *client_get_session_id(struct client *client);
 
 bool client_read(struct client *client);
 void client_input(struct client *client);
 
-void client_send_line(struct client *client, enum client_cmd_reply reply,
-		      const char *text);
+void client_notify_auth_ready(struct client *client);
+void client_notify_status(struct client *client, bool bad, const char *text);
+void client_notify_disconnect(struct client *client,
+			      enum client_disconnect_reason reason,
+			      const char *text);
+
 void client_send_raw_data(struct client *client, const void *data, size_t size);
 void client_send_raw(struct client *client, const char *data);
 
 void client_set_auth_waiting(struct client *client);
 void client_auth_send_challenge(struct client *client, const char *data);
-int client_auth_parse_response(struct client *client);
+void client_auth_parse_response(struct client *client);
 int client_auth_begin(struct client *client, const char *mech_name,
 		      const char *init_resp);
 bool client_check_plaintext_auth(struct client *client, bool pass_sent);
